;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; SCHEDULE.ASM - Kernel Scheduling Code for SimpleOS
;; (c) 2001 Breckin Loggins
;; This code is distributed under the terms of the GNU
;; General Public License

;; Function Exports
GLOBAL _scheduler		; Invoked by IRQ 0
GLOBAL _schedAdd		; Add a task to the run queue
;GLOBAL _schedDel		; Delete a task from the run queue

;; Function Imports
EXTERN _vidWriteString
EXTERN _printReg

EXTERN SendEOI

;; Public Functions

;; FUNCTION: _scheduler
;; ARGS:     none
;; PURPOSE: Simple round-robin scheduler that rotates through tasks in the 
;;	    run queue

_scheduler:
	; The scheduler is complicated because there are several possibilities 
	; and branch paths.  The first possibility is that the run queue is 
	; empty, this will only occur if current task is set to 0xff.  If this
	; is the case then there are further two possibilites, one is that the 
	; runqueue is truly empty, in which case we exit back to the task we 
	; interrupted (with a simple ret), the second case is that the first
	; task has been added to the queue but hasn't executed yet.  We need
	; some way to detect this.  The solution is that the first task will
	; ALWAYS occupy the first spot in the runqueue.  So we test this spot, 
	; and if it's zero we just exit, if there's something there it means
	; we have a brand new first task that hasn't executed yet, so we set
	; the current task to 0x0 and JMP to it.
	;
	; Once that is done we have another situation, that of there being only
	; one task in the queue.  We can test this by doing a loop through the
	; queue and seeing how many tasks we find, if there's only one, then 
	; just exit, don't JMP.  If we find more than one, it means we need to
	; find the next non-zero entry in the run queue, set it to the current
	; task, then JMP to it.  This algorithm will take care of all special 
	; cases in the scheduler.
	
	; First, xor all regs and set some appropriate values
	xor eax, eax
	xor ebx, ebx
	xor ecx, ecx
	
	mov eax, runqueue		; Address of runqueue
	
	; OK so step one, check currenttask for 0xff
	cmp byte [currenttask], 0xff
	jne .nextcheck			; There is at least one task, so skip
					; the rest of this algorithm
	
	; We know if we got here that currenttask is 0xff, but we don't know
	; if there is a task that hasn't been jmpd to yet, or if there are no
	; tasks, find out which one it is
	cmp word [eax], 0		; Is there a task waiting?
	jz .jend			; Nope, just leave
	jmp .njend
	
.jend:
	jmp .end
	
.njend:
	; Now we know that we have a new task that's the very first task and is
	; waiting to be executed for the first time, so set currenttask to 0
	; and jump to this task
	mov byte [currenttask], 0
	mov bx, [eax]			; Get the task's TSS
	mov word [thistss+4], bx	; Put it in the TSS descriptor
	
	call SendEOI
	jmp dword far [thistss]		; JMP!
	jmp .end			; When we return from here, leave from
					; this spot
.nextcheck:
	; OK, so at this point we know that there is at least 1 task that has
	; executed at least once, so we search the runqueue starting at
	; [currenttask] for the next available task to execute, if we get 
	; back to [currenttask] without finding one, we know there is only one
	; task in the queue so just exit
	
	; Get the runqueue offset from curren task
	xor eax, eax
	mov al, [currenttask]		; Get this task
	mov ebx, 2			
	mul ebx				; Now eax contains the offset, just add
					; the address of runqueue to it
	add eax, runqueue
	
.continue:
	add eax, 2			; Move one down
	cmp eax, runqueue + 200		; Are we at the end?
	jae .reset			; YES! go back to the beginning
	
.nextstep:
	; OK now we have an entry at least one away from currenttask, first
	; check to see if it's non zero
	cmp word [eax], 0
	jz .continue			; It is zero, keep going
	
	; We know it's non zero, check to see if it's currentask

	; Get the offset into the runqueue
	mov ebx, eax			; Save eax
	sub eax, runqueue		; Subtract the start of runqueue
	mov ecx, 2
	div ecx				; Now al contains the task number
	
	mov cl, [currenttask]		; Get the current task
	cmp al, cl		
	je .end				; It is, so we're the only task in the
					; queue, exit
	mov eax, ebx			; Restore eax
	
	; Now we have a nonzero task in the runqueue which isn't our current
	; task, so set it as the current task and JMP to it
	mov ebx, eax			; Save eax
	sub eax, runqueue
	mov ecx, 2
	div ecx
	mov byte [currenttask], al
	
	
	mov eax, ebx			; Restore eax
	
	mov bx, [eax]			; Get the task
	
	mov word [thistss+4], bx	; put it in thistss
	
	call SendEOI
	jmp dword far [thistss]		; JUMP!
	jmp .end			; When we return, just fall through to
					; the end
	
.reset
	mov eax, runqueue		; Go back to the beginning
	jmp .nextstep

.end:	
	call SendEOI
	ret
; End _scheduler

;; FUNCTION: _schedAdd
;; ARGS:     Push the TSS to the stack (0x20, 0x28, 0x30, etc...)
;; PURPOSE: Adds the selected TSS to the runqueue by running through the 
;;	    runqueue and adding the task to the first zero entry it finds
_schedAdd:
	push eax
	push ebx
	
	xor eax, eax
	xor ebx, ebx
	
	; Put the new task into bx
	mov bx, [esp+12]		
	
	; Put the starting address of the runqueue into eax
	mov eax, runqueue
	
.continue:
	; See if this address is zero
	cmp word [eax], 0
	jz .found			; It is, write it out
	
	; If we got here, we know that this particular entry isn't zero
	add eax, 2			; Increment runqueue ptr by 2 bytes
	
	; Now we need to check to make sure we're not at the end of the 
	; queue
	cmp eax, runqueue + 200
	jae .fatal
	
	; At this point we're sure we aren't at the end of the run queue, so
	; try again
	jmp .continue
	
.fatal:
	; If we got here we know that we're out of tasks, so panic and hlt the
	; processor
	push dword fatalerror
	call _vidWriteString
	cli
	hlt
	

.found:
	mov [eax], bx			; Enter the new task in the run queue
	
.end:
	pop ebx
	pop eax
	ret 4
; End _schedAdd

;; Helper Functions


;; Data Area

;; This is the runqueue, right now it's just an array of 100 possible task 
;; entries.  This will be replaced by a REAL queue soon :)
runqueue times 110 dw 0		;; Allocate 200 bytes for the runqueue

;; This is the current task (offset into the queue)
currenttask	db 0xff		; This is how we know we aren't fully
				; set up yet

;; This is the TSS we jump to (six bytes)
thistss		dd 0
		dd 0

fatalerror	db 'FATAL ERROR: Out of spaces in the task queue!', 0
